(*
Copyright 2009 Christian Sternagel, René Thiemann, Sarah Winkler, Harald Zankl

This file is part of IsaFoR/CeTA.

IsaFoR/CeTA is free software: you can redistribute it and/or modify it under the
terms of the GNU Lesser General Public License as published by the Free Software
Foundation, either version 3 of the License, or (at your option) any later
version.

IsaFoR/CeTA is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License along
with IsaFoR/CeTA. If not, see <http://www.gnu.org/licenses/>.
*)
theory DpTransImpl imports TermImpl DpFramework begin

types ('f,'v)dp = "('f shp,'v)rule" 
types ('f,'v)dpL = "('f,'v)dp list"


fun ID_list :: "('f,'v)trsL \<Rightarrow> ('f shp,'v)trsL"
where "ID_list R = map (\<lambda> lr. (plain (fst lr), plain (snd lr))) R"

lemma ID_list[simp] : "set (ID_list R) = ID (set R)"
proof (induct R, simp add: ID_def)
  case (Cons lr R)
  have "ID (set (lr # R)) = {(s,t). \<exists> l r. s = plain l \<and> t = plain r \<and> (l, r) \<in> set (lr # R)}" by (simp add: ID_def)
  also have "\<dots> = {(s,t). \<exists> l r. s = plain l \<and> t = plain r \<and> (l, r) \<in> set R} \<union> {(s,t). \<exists> l r. s = plain l \<and> t = plain r \<and> (l, r) = lr}" by auto
  also have "\<dots> = set (ID_list R) \<union> {(s,t). \<exists> l r. s = plain l \<and> t = plain r \<and> (l, r) = lr}" using Cons by (auto simp: ID_def)
  finally show ?case by (cases lr, auto)
qed

declare ID_list.simps[simp del]

definition check_wf_trs :: "('f::shows,'v::shows)trsL \<Rightarrow> shows monad" 
where "check_wf_trs R \<equiv>  (
  checkAllM (\<lambda>lr. checkB (\<not> is_var (fst lr)) (''the left-hand side of rule '' +#+ shows_rule lr +@+ '' is a variable'' +#+ shows_nl)) R
\<guillemotright> checkAllM (\<lambda>lr.
  (checkSubset (vars_term_list(snd lr)) (vars_term_list(fst lr))
   \<leftarrow> (\<lambda>x. ''the variable '' +#+ shows x +@+ '' does only occur in the right-hand side of rule '' +#+ shows_rule lr)
  )) R
  \<leftarrow> (\<lambda>e. ''the TRS is not well-formed'' +#+ shows_nl +@+ e)
)"

lemma check_wf_trs[simp]: "isOK(check_wf_trs R) = wf_trs (set R)" (is "?l = ?r")
proof
  assume ?l
  thus ?r unfolding check_wf_trs_def wf_trs_def
  proof (simp, clarify)
    fix l r
    assume mem: "(l,r) \<in> set R"
      and "\<forall> lr \<in> set R. \<not> is_var (fst lr)"
      and c2: "\<forall> lr \<in> set R. vars_term(snd lr) \<subseteq> vars_term(fst lr)"
    hence c1: "\<not> is_var (fst (l,r))" by auto
    from mem c2 have "vars_term(snd (l,r)) \<subseteq> vars_term(fst (l,r))" by force
    with c1 show "(\<exists>f ts. l = Fun f ts) \<and> vars_term r \<subseteq> vars_term l" 
      by (auto)
  qed
next
  assume ?r
  hence "\<forall> lr \<in> (set R). \<not> is_var (fst lr)" unfolding wf_trs_def by force
  hence one: "\<And> s. isOK (checkAllM (\<lambda> lr. checkB (\<not> is_var (fst lr)) s) R)" by auto
  from `?r` have "\<forall> lr \<in> set R. set (vars_term_list(snd lr)) \<subseteq> set(vars_term_list(fst lr))" unfolding wf_trs_def by force
  hence two: "isOK (checkAllM (\<lambda> lr. (checkSubset (vars_term_list(snd lr)) (vars_term_list(fst lr)))) R)" by auto 
  from one two show ?l unfolding check_wf_trs_def by auto
qed

definition DP_list :: "('f,'v)trsL \<Rightarrow> ('f shp,'v)trsL"
where "DP_list R \<equiv> let D = defs_list R in concat (map (\<lambda> lr. let l = fst lr; s = sharp l in map (\<lambda> u. (s, sharp u)) (filter (\<lambda> u. \<not> l \<rhd> u \<and> \<not> is_var u \<and> the (root u) mem D) (supteq_list (snd lr)))) R)"

lemma DP_list[simp]: "set (DP_list R) = DP (set R)"
proof -
  let ?p = "\<lambda> s t l r u. s = sharp l \<and> t = sharp u \<and> (l,r) \<in> set R \<and> \<not> l \<rhd> u \<and> r \<unrhd> u \<and> \<not> is_var u \<and> defined (set R) (the (root u))"
  let ?q = "\<lambda> s t l r h us.  s = sharp l \<and> t = sharp (Fun h us) \<and> (l, r) \<in> set R \<and> r \<unrhd> Fun h us \<and> defined (set R) h \<and> \<not> l \<rhd> Fun h us"
  have "set (DP_list R) = (\<Union>(l,r) \<in> set R. ((\<lambda>u. (sharp l, sharp u)) ` {u. r \<unrhd> u \<and> (l \<unrhd> u \<longrightarrow> l = u) \<and> \<not> is_var u \<and> defined (set R) (the (root u))}))"
    unfolding DP_list_def by (auto simp: Let_def defs_list_sound[symmetric])
  also have "\<dots> = {(s,t). \<exists> l r u. ?p s t l r u}" by auto
  also have "\<dots> = {(s,t). \<exists> l r h us. ?q s t l r h us}" 
  proof -
    {
      fix s t l r
      have "(\<exists> u. ?p s t l r u) = (\<exists> h us. ?q s t l r h us)" (is "?l = ?r")
      proof
        assume ?l
        then obtain u where "?p s t l r u" ..
        then obtain h us where "?q s t l r h us" by (cases u, auto)
        thus ?r by blast
      next
        assume ?r 
        then obtain h us where "?q s t l r h us" by auto
        hence "?p s t l r (Fun h us)" by auto
        thus ?l by blast
      qed          
    }
    thus ?thesis by simp
  qed
  also have "\<dots> = DP (set R)" unfolding DP_def by simp
  finally show ?thesis .
qed
      
definition check_DP :: "('f::shows,'v::shows)trsL \<Rightarrow> ('f,'v)dpL \<Rightarrow> shows monad"
where "check_DP r P \<equiv> (
  check_wf_trs r   
\<guillemotright> checkSubset (DP_list r) P 
  \<leftarrow> (\<lambda>dp. ''the DP '' +#+ shows_rule dp +@+ '' does not appear in the DP problem'' +#+ shows_nl)
) \<leftarrow> (\<lambda>s.  ''the DP-transformation is not applied correctly.'' +#+ shows_nl +@+ s)
"

lemma check_DP : assumes check: "isOK(check_DP r P)"
  shows "finite_dpp (set P, ID (set r)) \<Longrightarrow> SN (rstep (set r))"
proof -
  assume fin: "finite_dpp (set P, ID (set r))"
  show "SN (rstep (set r))"
  proof -
    from check have wf: "wf_trs (set r)" 
      and dp: "Trs.DP (set r) \<subseteq> set P" 
      unfolding check_DP_def by auto
    let ?DP = "Trs.DP (set r)"
    from fin have "finite_dpp (set P, ID (set r))"  by simp
    hence fin2: "finite_dpp (set P \<inter> ?DP, ID (set r))" by (unfold finite_dpp_def, auto)
    from dp have "?DP = set P \<inter> ?DP" by auto
    with fin2 have "finite_dpp (initial_dpp (set r))" by simp
    with wf show "?thesis" by (rule finite_ini_imp_SN)
  qed
qed

definition check_not_wf_trs :: "('f::shows,'v::shows)trsL \<Rightarrow> shows monad"
  where "check_not_wf_trs R \<equiv> checkB (\<not> isOK (check_wf_trs R)) (''The TRS is well formed'' +#+ shows_nl)"
  

lemma check_not_wf_trs : assumes check: "isOK(check_not_wf_trs R)"
  shows "\<not> SN (rstep (set R))"
proof -
  from assms have "\<not> wf_trs (set R)" unfolding check_not_wf_trs_def by auto
  with SN_rstep_imp_wf_trs show ?thesis by auto
qed
  

end
